"""Parse DFD source text in our DSL"""

import os.path
import re
from typing import Callable, Tuple

from . import dfd_dot_templates as TMPL
from . import model


def check(statements: model.Statements) -> dict[str, model.Item]:
    """Check that all statement make sense"""

    # collect items
    items_by_name: dict[str, model.Item] = {}
    for statement in statements:
        error_prefix = model.mk_err_prefix_from(statement.source)
        match statement:
            case model.Item() as item: pass
            case _: continue
        name = item.name

        # make sure there are no duplicates
        if name not in items_by_name:
            items_by_name[name] = item
            continue

        other = items_by_name[name]
        other_text = model.pack(other.source.text)
        raise model.DfdException(
            f'{error_prefix}Name "{name}" already exists '
            f'at line {other.source.line_nr+1}: {other_text}')

    # check references and values of connections
    for statement in statements:
        error_prefix = model.mk_err_prefix_from(statement.source)
        match statement:
            case model.Connection() as conn: pass
            case _: continue
        nb_stars = 0
        for point in conn.src, conn.dst:
            if point == "*":
                nb_stars += 1
            if point != "*":
                if point not in items_by_name:
                    raise model.DfdException(
                        f'{error_prefix}Connection "{conn.type}" links to "{point}", '
                        f'which is not defined')
                if items_by_name[point].type == model.CONTROL and conn.type != model.SIGNAL:
                    raise model.DfdException(
                        f'{error_prefix}Connection to {model.CONTROL} "{point}" is '
                        f'of type "{conn.type}", however only connections of type '
                        f'"{model.SIGNAL}" are allowed')

        if nb_stars == 2:
            raise model.DfdException(
                f'{error_prefix}Connection "{conn.type}" may not link to two '
                f'stars')

    # check references of frames
    framed_items: set[str] = set()
    for statement in statements:
        error_prefix = model.mk_err_prefix_from(statement.source)
        match statement:
            case model.Frame() as frame: pass
            case _: continue
        if not frame.items:
            raise model.DfdException(
                f'{error_prefix}Frame is empty')
        for name in frame.items:
            if name not in items_by_name:
                raise model.DfdException(
                    f'{error_prefix}Frame includes "{name}", '
                    f'which is not defined')
            if name in framed_items:
                raise model.DfdException(
                    f'{error_prefix}Item "{name}", '
                    f'is in multiple frames')
            framed_items.add(name)

    return items_by_name


def parse(source_lines: model.SourceLines, debug: bool = False,
         ) -> tuple[model.Statements, model.GraphDependencies]:
    """Parse the DFD source text as list of statements"""

    statements: model.Statements = []
    dependencies: model.GraphDependencies = []

    for n, source in enumerate(source_lines):
        src_line = source.text

        src_line = src_line.strip()
        if not src_line or src_line.startswith('#'): continue
        error_prefix = model.mk_err_prefix_from(source)

        # syntactic sugars may rewrite the line
        line = apply_syntactic_sugars(src_line)
        source.text = line  # fixup text

        word = source.text.split()[0]

        f: Callable[[model.SourceLine], model.Statement] = {
            model.STYLE: parse_style,
            model.PROCESS: parse_process,
            model.CONTROL: parse_control,
            model.ENTITY: parse_entity,
            model.STORE: parse_store,
            model.NONE: parse_none,
            model.CHANNEL: parse_channel,

            model.FLOW: parse_flow,
            model.CFLOW: parse_cflow,
            model.BFLOW: parse_bflow,
            model.UFLOW: parse_uflow,
            model.SIGNAL: parse_signal,

            model.FLOW+'?': parse_flow_q,
            model.CFLOW+'?': parse_cflow_q,
            model.BFLOW+'?': parse_bflow_q,
            model.UFLOW+'?': parse_uflow_q,
            model.SIGNAL+'?': parse_signal_q,

            'flow.r': parse_flow_r,
            'cflow.r': parse_cflow_r,
            'signal.r': parse_signal_r,

            'flow.r?': parse_flow_r_q,
            'cflow.r?': parse_cflow_r_q,
            'signal.r?': parse_signal_r_q,

            model.FRAME: parse_frame,

        }.get(word)

        if f is None:
            raise model.DfdException(f'{error_prefix}Unrecognized keyword '
                                     f'"{word}"')

        try:
            statement = f(source)
        except model.DfdException as e:
            raise model.DfdException(f'{error_prefix}{e}"')

        match statement:
            case model.Item() as item:
                parse_item_external(item, dependencies)
                item.text = item.text or item.name

        match statement:
            case model.Drawable() as drawable:
                parse_drawable_attrs(drawable)

        statements.append(statement)

    if debug:
        for s in statements: print(model.repr(s))
    return statements, dependencies


def split_args(dfd_line: str, n: int, last_is_optional: bool=False) -> list[str]:
    """Split DFD line into n (possibly n-1) tokens"""

    terms = dfd_line.split(maxsplit=n)
    if len(terms)-1 == n-1 and last_is_optional:
        terms.append(None)

    if len(terms)-1 != n:
        if not last_is_optional:
            raise model.DfdException(f'Expected {n} argument(s)')
        else:
            raise model.DfdException(f'Expected {n-1} or {n} argument')

    return terms[1:]


def parse_item_name(name: str) -> Tuple[str, bool]:
    """If name ends with ?, make it hidable"""
    if name.endswith('?'):
        return name[:-1], True
    else:
        return name, False


def parse_style(source: model.SourceLine) -> model.Statement:
    """Parse style statement"""
    style, value = split_args(source.text, 2, True)
    return model.Style(source, style, value)


def parse_process(source: model.SourceLine) -> model.Statement:
    """Parse process statement"""
    name, text = split_args(source.text, 2, True)
    name, hidable = parse_item_name(name)
    return model.Item(source, model.PROCESS, text, "", name, hidable)


def parse_control(source: model.SourceLine) -> model.Statement:
    """Parse control statement"""
    name, text = split_args(source.text, 2, True)
    name, hidable = parse_item_name(name)
    return model.Item(source, model.CONTROL, text, "", name, hidable)


def parse_entity(source: model.SourceLine) -> model.Statement:
    """Parse entity statement"""
    name, text = split_args(source.text, 2, True)
    name, hidable = parse_item_name(name)
    return model.Item(source, model.ENTITY, text, "", name, hidable)


def parse_store(source: model.SourceLine) -> model.Statement:
    """Parse store statement"""
    name, text = split_args(source.text, 2, True)
    name, hidable = parse_item_name(name)
    return model.Item(source, model.STORE, text, "", name, hidable)


def parse_none(source: model.SourceLine) -> model.Statement:
    """Parse none statement"""
    name, text = split_args(source.text, 2, True)
    name, hidable = parse_item_name(name)
    return model.Item(source, model.NONE, text, "", name, hidable)


def parse_channel(source: model.SourceLine) -> model.Statement:
    """Parse channel statement"""
    name, text = split_args(source.text, 2, True)
    name, hidable = parse_item_name(name)
    return model.Item(source, model.CHANNEL, text, "", name, hidable)


def parse_flow(source: model.SourceLine) -> model.Statement:
    """Parse directional flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.FLOW, text, "", src, dst)


def parse_flow_r(source: model.SourceLine) -> model.Statement:
    """Parse directional reversed flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.FLOW, text, "", src, dst, True)


def parse_cflow(source: model.SourceLine) -> model.Statement:
    """Parse continuous flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.CFLOW, text, "", src, dst)

def parse_cflow_r(source: model.SourceLine) -> model.Statement:
    """Parse continuous flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.CFLOW, text, "", src, dst, True)

def parse_bflow(source: model.SourceLine) -> model.Statement:
    """Parse bidirectional flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.BFLOW, text, "", src, dst)


def parse_uflow(source: model.SourceLine) -> model.Statement:
    """Parse undirected flow flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.UFLOW, text, "", src, dst)


def parse_signal(source: model.SourceLine) -> model.Statement:
    """Parse signal statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.SIGNAL, text, "", src, dst)


def parse_signal_r(source: model.SourceLine) -> model.Statement:
    """Parse reversed signal statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.SIGNAL, text, "", src, dst, True)

def parse_flow_q(source: model.SourceLine) -> model.Statement:
    """Parse directional flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.FLOW, text, "", src, dst,
                            relaxed=True)


def parse_flow_r_q(source: model.SourceLine) -> model.Statement:
    """Parse directional reversed flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.FLOW, text, "", src, dst, True,
                            relaxed=True)


def parse_cflow_q(source: model.SourceLine) -> model.Statement:
    """Parse continuous flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.CFLOW, text, "", src, dst,
                            relaxed=True)

def parse_cflow_r_q(source: model.SourceLine) -> model.Statement:
    """Parse continuous flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.CFLOW, text, "", src, dst, True,
                            relaxed=True)


def parse_bflow_q(source: model.SourceLine) -> model.Statement:
    """Parse bidirectional flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.BFLOW, text, "", src, dst,
                            relaxed=True)


def parse_uflow_q(source: model.SourceLine) -> model.Statement:
    """Parse undirected flow flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.UFLOW, text, "", src, dst,
                            relaxed=True)


def parse_signal_q(source: model.SourceLine) -> model.Statement:
    """Parse signal statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.SIGNAL, text, "", src, dst,
                            relaxed=True)


def parse_signal_r_q(source: model.SourceLine) -> model.Statement:
    """Parse reversed signal statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.SIGNAL, text, "", src, dst, True,
                            relaxed=True)


def apply_syntactic_sugars(src_line: str) -> str:
    terms = src_line.split()
    if len(terms) < 3:
        return src_line

    op = terms[1]

    def fmt(verb: str, args: list[str], swap: bool = False) -> str:
        array = [verb] + args
        del array[2]  # remove arrow
        if swap:
            array[1], array[2] = array[2], array[1]
        return '\t'.join(array)

    new_line = ''
    if re.fullmatch(r'-+>[?]?', op):
        q = '?' if op.endswith('?') else ''
        parts = src_line.split(maxsplit=3)
        new_line = fmt('flow'+q, parts)
    elif re.fullmatch(r'<-+[?]?', op):
        q = '?' if op.endswith('?') else ''
        parts = src_line.split(maxsplit=3)
        new_line = fmt('flow.r'+q, parts)#, swap=True)

    if re.fullmatch(r'-+>>[?]?', op):
        q = '?' if op.endswith('?') else ''
        parts = src_line.split(maxsplit=3)
        new_line = fmt('cflow'+q, parts)
    elif re.fullmatch(r'<<-+[?]?', op):
        q = '?' if op.endswith('?') else ''
        parts = src_line.split(maxsplit=3)
        new_line = fmt('cflow.r'+q, parts)#, swap=True)

    elif re.fullmatch(r'<-+>[?]?', op):
        q = '?' if op.endswith('?') else ''
        parts = src_line.split(maxsplit=3)
        new_line = fmt('bflow'+q, parts)

    elif re.fullmatch(r'--+[?]?', op):
        q = '?' if op.endswith('?') else ''
        parts = src_line.split(maxsplit=3)
        new_line = fmt('uflow'+q, parts)

    elif re.fullmatch(r':+>[?]?', op):
        q = '?' if op.endswith('?') else ''
        parts = src_line.split(maxsplit=3)
        new_line = fmt('signal'+q, parts)
    elif re.fullmatch(r'<:+[?]?', op):
        q = '?' if op.endswith('?') else ''
        parts = src_line.split(maxsplit=3)
        new_line = fmt('signal.r'+q, parts)#, swap=True)

    if new_line:
        return new_line
    else:
        return src_line


def parse_drawable_attrs(drawable: model.Drawable) -> None:
    if drawable.text and drawable.text.startswith('['):
        parts = drawable.text[1:].split(']', 1)
        drawable.attrs = parts[0]
        drawable.text = parts[1].strip()

        match drawable:
            case model.Item() as item:
                item.text = item.text or item.name


def parse_item_external(item: model.Item, dependencies: model.GraphDependencies) -> None:
    parts = item.name.split(':', 1)
    if len(parts) > 1:
        item.attrs = TMPL.ITEM_EXTERNAL_ATTRS
        if parts[-1]:
            item.name = parts[-1]
        else:
            item.name = parts[-2]

        if item.name.startswith('#'):
            item.name = item.name[1:]
        else:
            item.name = os.path.splitext(item.name)[0]

        if not item.text:
            item.text = item.name

        dependency = model.GraphDependency(parts[0],
                                           parts[1] or None,
                                           item.type,
                                           item.source)
        dependencies.append(dependency)


def parse_frame(source: model.SourceLine) -> model.Statement:
    """Parse frame statement"""
    parts = source.text.split('=', maxsplit=1)
    if len(parts) == 1:
        text = ''
    else:
        text = parts[1].strip()

    items = parts[0].split()[1:]
    type = ''  # so far there is only one type of frame
    attrs = 'style=dashed'
    return model.Frame(source, type, text, attrs, items)



"""Command-line DFD diagram generator. Converts a textual description
into a graphic file.

-----

See https://github.com/pbauermeister/dfd for information, syntax and
examples.

-----

This module parses the commandline args, prepares the I/Os and calls the stuff.
"""

import argparse
import io
import os
import re
import sys
import tempfile
from typing import TextIO

import pkg_resources

from . import dfd, dot, markdown, model
from .error import print_error

try:
    VERSION = pkg_resources.require("data-flow-diagram")[0].version
except pkg_resources.DistributionNotFound:
    VERSION = 'undefined'


def parse_args() -> argparse.Namespace:
    description, epilog = [each.strip() for each in __doc__.split('-----')[:2]]

    parser = argparse.ArgumentParser(description=description, epilog=epilog)

    parser.add_argument('INPUT_FILE',
                        action='store',
                        default=None, nargs='?',
                        help='UML sequence input file; '
                        'if omitted, stdin is used')

    parser.add_argument('--output-file', '-o',
                        required=False,
                        help='output file name; pass \'-\' to use stdout; '
                        'if omitted, use INPUT_FILE base name with \'.svg\' '
                        'extension, or stdout')

    parser.add_argument('--markdown', '-m',
                        action='store_true',
                        help='consider snippets between opening marker: '
                        '```data-flow-diagram OUTFILE, and closing marker: ``` '
                        'allowing to generate all diagrams contained in an '
                        'INPUT_FILE that is a markdown file');

    parser.add_argument('--format', '-f',
                        required=False,
                        default='svg',
                        help='output format: gif, jpg, tiff, bmp, pnm, eps, '
                        'pdf, svg (any supported by Graphviz); default is svg')

    parser.add_argument('--percent-zoom', '-p',
                        required=False,
                        default=100, type=int,
                        help='magnification percentage; default is 100')

    parser.add_argument('--background-color', '-b',
                        required=False,
                        default='white',
                        help='background color name (including \'none\' for'
                        ' transparent) in web color notation; see'
                        ' https://developer.mozilla.org/en-US/docs/Web/CSS/color_value'
                        ' for a list of valid names; default is white')

    parser.add_argument('--no-graph-title',
                        action='store_true',
                        default=False,
                        help='suppress graph title')

    parser.add_argument('--no-check-dependencies',
                        action='store_true',
                        default=False,
                        help='suppress dependencies checking')

    parser.add_argument('--debug',
                        action='store_true',
                        default=False,
                        help='emit debug messages')

    parser.add_argument('--version', '-V',
                        action='store_true',
                        help='print the version and exit')

    return parser.parse_args()


def handle_markdown_source(options: model.Options, provenance: str,
                           input_fp: TextIO) -> None:
    text = input_fp.read()
    snippets = markdown.extract_snippets(text)
    markdown.check_snippets_unicity(provenance, snippets)
    snippets_params = markdown.make_snippets_params(provenance, snippets)
    for params in snippets_params:
        dfd.build(params.root, params.input_fp.read(), params.file_name,
                  options, snippet_by_name=params.snippet_by_name)
        print(f'{sys.argv[0]}: generated {params.file_name}', file=sys.stderr)


def handle_dfd_source(options: model.Options, provenance: str,
                      input_fp: TextIO, output_path: str) -> None:
    root = model.SourceLine("", provenance, None, None)
    if output_path == '-':
        # output to stdout
        with tempfile.TemporaryDirectory() as d:
            path = os.path.join(d, 'file.svg')
            dfd.build(root, input_fp.read(), path, options)
            with open(path) as f:
                print(f.read())
    else:
        # output to file
        dfd.build(root, input_fp.read(), output_path, options)


def run(args: argparse.Namespace) -> None:
    # adjust input
    if args.INPUT_FILE is None:
        input_fp = sys.stdin
        provenance = '<stdin>'
    else:
        input_fp = open(args.INPUT_FILE)
        provenance = f'<file:{args.INPUT_FILE}>'

    options = model.Options(
        args.format,
        args.percent_zoom,
        args.background_color,
        args.no_graph_title,
        args.no_check_dependencies,
        args.debug,
    )

    # markdown source
    if args.markdown:
        handle_markdown_source(options, provenance, input_fp)
        return

    # adjust output
    if args.output_file is None:
        if args.INPUT_FILE is not None:
            basename = os.path.splitext(args.INPUT_FILE)[0]
            output_path = basename + '.' + args.format
        else:
            output_path = '-'
    else:
        output_path = args.output_file

    # DFD source
    handle_dfd_source(options, provenance, input_fp, output_path)


def main() -> None:
    """Entry point for the application script"""

    dot.check_installed()

    args = parse_args()
    if args.version:
        print('data-flow-diagram', VERSION)
        sys.exit(0)

    try:
        run(args)
    except model.DfdException as e:
        text = f'ERROR: {e}'
        print_error(text)
        sys.exit(1)



"""Internal data model"""

from __future__ import annotations

import dataclasses
import json
from dataclasses import dataclass
from typing import Any

from . import config


def repr(o: Any) -> str:
    name: str = o.__class__.__name__
    val: str = json.dumps(dataclasses.asdict(o), indent='  ')
    return f'{name} {val}'


@dataclass
class Base:
    def __repr__(self) -> str:
        return (self.__class__.__name__ + ' '
                + json.dumps(dataclasses.asdict(self), indent='  '))


@dataclass
class Snippet(Base):
    text: str
    name: str
    output: str
    line_nr: int


@dataclass
class SourceLine(Base):
    text: str  # after pre-processor
    raw_text: str
    parent: SourceLine  # https://stackoverflow.com/a/62521947
    line_nr: int
    is_container: bool = False


@dataclass
class Statement(Base):
    source: SourceLine


@dataclass
class Drawable(Statement):
    type: str
    text: str
    attrs: str


@dataclass
class Item(Drawable):
    name: str
    hidable: bool


@dataclass
class Connection(Drawable):
    src: str
    dst: str
    reversed: bool = False
    relaxed: bool = False


@dataclass
class Frame(Drawable):
    items: list[str]


@dataclass
class Style(Statement):
    style: str
    value: Any = None


STYLE   = 'style'

PROCESS = 'process'
CONTROL = 'control'
ENTITY  = 'entity'
STORE   = 'store'
CHANNEL = 'channel'
NONE    = 'none'

FLOW    = 'flow'
BFLOW   = 'bflow'
CFLOW   = 'cflow'
UFLOW   = 'uflow'
SIGNAL  = 'signal'

FRAME   = 'frame'


@dataclass
class GraphOptions:
    is_vertical: bool = False
    is_context: bool = False
    item_text_width = config.DEFAULT_ITEM_TEXT_WIDTH
    connection_text_width = config.DEFAULT_CONNECTION_TEXT_WIDTH


def pack(src_line: str) -> str:
    if src_line is None:
        return '<none>'
    return ' '.join(src_line.split())


def mk_err_prefix_from(src: SourceLine) -> str:

    def _add_to_stack(stack: list[str], src: SourceLine) -> None:
        if src.line_nr is None:
            stack += [f'  {pack(src.raw_text)}']
        else:
            if src.parent and src.parent.is_container:
                nr = src.parent.line_nr+1
                delta = src.line_nr+1
                final = nr + delta
                stack += [f'  line {final}: {pack(src.raw_text)}']
            else:
                nr = src.line_nr+1
                stack += [f'  line {nr}: {pack(src.raw_text)}']
        if src.parent:
            _add_to_stack(stack, src.parent)

    stack: list[str] = ['(most recent first)']
    _add_to_stack(stack, src)
    stack += ['']
    return '\n'.join(stack) + 'Error: '


class DfdException(Exception):
    pass


# Handy type aliases
Snippets = list[Snippet]
SourceLines = list[SourceLine]
Statements = list[Statement]
SnippetByName = dict[str, Snippet]


@dataclass
class Options:
    format: str
    percent_zoom: int
    background_color: str
    no_graph_title: bool
    no_check_dependencies: bool
    debug: bool


@dataclass
class GraphDependency:
    to_graph: str
    to_item: str
    to_type: str
    source: SourceLine

GraphDependencies = list[GraphDependency]


import sys

from . import model, scanner, parser


def check(dependencies: model.GraphDependencies, snippet_by_name: model.SnippetByName,
          options: model.Options) -> None:

    snippet_by_name = snippet_by_name or {}
    errors: list[str] = []
    for dep in dependencies:
        prefix = model.mk_err_prefix_from(dep.source)

        # load source text
        if dep.to_graph.startswith('#'):
            # from snippet
            name = dep.to_graph[1:]
            if name not in snippet_by_name:
                errors.append(f'{prefix}Referring to unknown snippet "{name}"')
                continue
            text = snippet_by_name[name].text
            what = 'snippet'
        else:
            # from file
            name = dep.to_graph
            try:
                with open(name, encoding='utf-8') as f:
                    text = f.read()
            except FileNotFoundError as e:
                if name in snippet_by_name:
                    errors.append(f'{prefix}{e}. Did you mean "#{name}" ?')
                else:
                    errors.append(f'{prefix}{e}')
                continue
            what = 'file'

        # if only graph is targetted, we're done
        if dep.to_item is None:
            if dep.to_type != model.NONE:
                errors.append(f'{prefix}A whole graph may only be referred to '
                              f'by an item of type "{model.NONE}", and not '
                              f'"{dep.to_type}"')
            continue

        # scan and parse
        lines = scanner.scan(dep.source, text, snippet_by_name, options.debug)
        statements, _ = parser.parse(lines, options.debug)

        # find name
        item = find_item(dep.to_item, statements)
        if item:
            if dep.to_type != item.type:
                errors.append(f'{prefix}Referred item "{dep.to_item}" is of '
                              f'type "{item.type}", but is referred to as '
                              f'type "{dep.to_type}"')

            continue  # Found!

        errors.append(f'{prefix}Referring to unknown item name "{dep.to_item}"'
                      f' of {what} "{name}"')

    if errors:
        errors.insert(0, 'Dependency error(s) found:')
        raise model.DfdException('\n\n'.join(errors))


def find_item(name: str, statements: model.Statements) -> model.Item:
    for statement in statements:
        match statement:
            case model.Item() as item:
                if item.name == name:
                    return statement
    return None



import io
import os
import re
from collections import Counter
from dataclasses import dataclass
from typing import TextIO

from . import model

@dataclass
class SnippetContext:
    root: model.SourceLine
    input_fp: TextIO
    file_name: str
    snippet_by_name: model.SnippetByName

SnippetContexts = list[SnippetContext]


def extract_snippets(text: str) -> model.Snippets:
    rx = re.compile(r'^```(?P<head>\s*)'
                    r'data-flow-diagram\s+'
                    r'(?P<output>.*?)\s*'
                    r'^(?P<src>.*?)^\s*```', re.DOTALL | re.M)

    return [
        model.Snippet(
            text=match['head'] + match['src'],
            name=os.path.splitext(match['output'])[0],
            output=match['output'],
            line_nr=len(text[:match.start()].splitlines()),
            )
        for match in rx.finditer(text)]
        # FIXME: have text a TextIO and find in streaming


def make_snippets_params(provenance: str, snippets: model.Snippets,
                        ) -> SnippetContexts:
    res: SnippetContexts = []
    snippet_by_name = {s.name:s for s in snippets}

    for snippet in snippets:
        # snippet w/o output, maybe just as includee
        if snippet.output.startswith('#'):
            continue

        # snippet with output
        input_fp = io.StringIO(snippet.text)
        snippet_provenance = f'{provenance}<snippet:{snippet.output}>'
        root = model.SourceLine("", snippet_provenance,
                                None, snippet.line_nr,
                                is_container=True)
        file_name = input_fp.name = snippet.output
        res.append(SnippetContext(root, input_fp, file_name, snippet_by_name))

    return res


def check_snippets_unicity(provenance: str, snippets: model.Snippets) -> None:
    snippet_names = [s.name for s in snippets]  # capture duplicates
    counts = Counter(snippet_names)
    multiples = {k:n for k, n in counts.items() if n>1}
    if multiples:
        root = model.SourceLine("", provenance, None, None)
        error_prefix = model.mk_err_prefix_from(root)
        raise model.DfdException(f'{error_prefix}Snippets defined multiple '
                                 f'times: {multiples}')



STORE = """
"{name}" [shape=none label=<
  <TABLE BORDER="0">
    <TR><TD BGCOLOR="black" WIDTH="6"></TD></TR>
    <TR><TD>{text}</TD></TR>
    <TR><TD BGCOLOR="black" WIDTH="6"></TD></TR>
  </TABLE>>]
""".strip()


CHANNEL = """
"{name}" [shape=none label=<
  <TABLE BORDER="0">
    <TR>
      <TD WIDTH="48"></TD>
      <TD BGCOLOR="black" WIDTH="0" PORT="x"><BR/><BR/></TD>
      <TD WIDTH="48"></TD>
    </TR>
    <TR>
      <TD COLSPAN="3">{text}</TD>
    </TR>
  </TABLE>>]
""".strip()


CHANNEL_HORIZONTAL = """
"{name}" [shape=none label=<
  <TABLE BORDER="0">
    <TR><TD BGCOLOR="black" PORT="x"></TD></TR>
    <TR><TD>{text}</TD></TR>
  </TABLE>>]
""".strip()


DOT_FONT_EDGE = 'fontname="times-italic" fontsize=10'
DOT_FONT_NODE = 'fontname="helvetica" fontsize=10'
DOT_FONT_GRAPH = 'fontname="helvetica" fontsize=9 fontcolor="#000060"'

DOT_GRAPH_TITLE = """graph[label="\n- {title} -" """ + DOT_FONT_GRAPH + """]"""

DOT = """
digraph D {{
  {graph_params}
  edge[color=gray """ + DOT_FONT_EDGE + """]
  node[""" + DOT_FONT_NODE + """]
  {block}
}}
""".strip()


GRAPH_PARAMS_CONTEXT_DIAGRAM = 'edge [len=2.25]'


ITEM_EXTERNAL_ATTRS = 'fillcolor=white color=grey fontcolor=grey'


"""Run the generation process"""

import os.path
import sys
import textwrap
from typing import Any, Optional

from . import config
from . import dfd_dot_templates as TMPL
from . import dot, model, parser, scanner, dependency_checker


def build(provenance: model.SourceLine, dfd_src: str, output_path: str,
          options: model.Options,
          snippet_by_name: model.SnippetByName = None) -> None:
    """Take a DFD source and build the final image or document"""
    lines = scanner.scan(provenance, dfd_src, snippet_by_name, options.debug)
    statements, dependencies = parser.parse(lines, options.debug)
    if dependencies and not options.no_check_dependencies:
        dependency_checker.check(dependencies, snippet_by_name, options)
    items_by_name = parser.check(statements)
    statements = filter_statements(statements)
    statements, graph_options = handle_options(statements)

    if options.no_graph_title:
        title = None
    else:
        title = os.path.splitext(output_path)[0]

    gen = Generator(graph_options)
    text = generate_dot(gen, title, statements, items_by_name)
    if options.debug:
        print(text)
    dot.generate_image(graph_options, text, output_path, options.format)


def wrap(text: str, cols: int) -> str:
    res: list[str] = []
    for each in text.strip().split('\\n'):
        res += textwrap.wrap(each, width=cols, break_long_words=False) or ['']
    return '\\n'.join(res)

class Generator:
    def __init__(self, graph_options: model.GraphOptions) -> None:
        self.lines: list[str] = []
        self.star_nr = 0
        self.frame_nr = 0
        self.graph_options = graph_options

    def append(self, line:str, statement: model.Statement) -> None:
        self.lines.append('')
        text = model.pack(statement.source.text)
        self.lines.append(f'/* {statement.source.line_nr}: {text} */')
        self.lines.append(line)

    def generate_item(self, item: model.Item) -> None:
        text = wrap(item.text, self.graph_options.item_text_width)
        attrs = item.attrs or ''
        match item.type:
            case model.PROCESS:
                if self.graph_options.is_context:
                    shape = 'circle'
                    fc = 'white'
                else:
                    shape = 'ellipse'
                    fc = '"#eeeeee"'
                line = (f'"{item.name}" [shape={shape} label="{text}" '
                        f'fillcolor={fc} style=filled {attrs}]')
            case model.CONTROL:
                fc = '"#eeeeee"'
                line = (f'"{item.name}" [shape=ellipse label="{text}" '
                        f'fillcolor={fc} style="filled,dashed" {attrs}]')
            case model.ENTITY:
                line = (f'"{item.name}" [shape=rectangle label="{text}" '
                        f'{attrs}]')
            case model.STORE:
                d = self._item_to_html_dict(item)
                line = TMPL.STORE.format(**d)
            case model.NONE:
                line = f'"{item.name}" [shape=none label="{text}" {attrs}]'
            case model.CHANNEL:
                d = self._item_to_html_dict(item)
                if self.graph_options.is_vertical:
                    line = TMPL.CHANNEL_HORIZONTAL.format(**d)
                else:
                    line = TMPL.CHANNEL.format(**d)
            case _:
                prefix = model.mk_err_prefix_from(item.source)
                raise model.DfdException(f'{prefix}Unsupported item type '
                                         f'"{item.type}"')
        self.append(line, item)

    def _item_to_html_dict(self, item: model.Item) -> dict[str, Any]:
        d = item.__dict__
        d['text'] = d['text'].replace('\\n', '<br/>')
        return d

    def generate_star(self, text: str) -> str:
        text = wrap(text, self.graph_options.item_text_width)
        star_name = f'__star_{self.star_nr}__'
        line = f'"{star_name}" [shape=none label="{text}" {TMPL.DOT_FONT_EDGE}]'
        self.lines.append(line)
        self.star_nr += 1
        return star_name

    def generate_connection(self, conn: model.Connection, src_item: model.Item,
                            dst_item: model.Item) -> None:
        text = conn.text or ''
        text = wrap(text, self.graph_options.connection_text_width)

        src_port = dst_port = ""

        if not src_item:
            src_name = self.generate_star(text)
            text = ''
        else:
            src_name = src_item.name
            if src_item.type == model.CHANNEL:
                src_port = ':x:c'

        if not dst_item:
            dst_name = self.generate_star(text)
            text = ''
        else:
            dst_name = dst_item.name
            if dst_item.type == model.CHANNEL:
                dst_port = ':x:c'

        attrs = f'label="{text}"'
        if conn.attrs: attrs += ' ' + conn.attrs

        match conn.type:
            case model.FLOW:
                if conn.reversed:
                    attrs += ' dir=back'
            case model.BFLOW:
                attrs += ' dir=both'
            case model.CFLOW:
                if conn.reversed:
                    attrs += ' dir=back'
                    attrs += ' arrowtail=normalnormal'
                else:
                    attrs += ' arrowhead=normalnormal'
            case model.UFLOW:
                attrs += ' dir=none'
            case model.SIGNAL:
                if conn.reversed:
                    attrs += ' dir=back'
                attrs += ' style=dashed'
            case _:
                prefix = model.mk_err_prefix_from(conn.source)
                raise model.DfdException(f'{prefix}Unsupported connection type '
                                         f'"{conn.type}"')
        if conn.relaxed:
            attrs += ' constraint=false'

        line = f'"{src_name}"{src_port} -> "{dst_name}"{dst_port} [{attrs}]'
        self.append(line, conn)

    def generate_style(self, style: model.Style) -> None:
        pass

    def generate_frame(self, frame: model.Frame) -> None:
        self.append(f'subgraph cluster_{self.frame_nr} {{', frame)
        self.frame_nr += 1

        self.lines.append(f'  label="{frame.text}"')
        if frame.attrs:
            self.lines.append(f'  {frame.attrs}')

        for item in frame.items:
            self.lines.append(f'  "{item}"')
        self.lines.append('}')

    def generate_dot_text(self, title: str) -> str:
        graph_params = []
        if self.graph_options.is_context:
            graph_params.append(TMPL.GRAPH_PARAMS_CONTEXT_DIAGRAM)

        if title:
            graph_params.append(TMPL.DOT_GRAPH_TITLE.format(title=title))

        if self.graph_options.is_vertical:
            graph_params.append('rankdir=TB')
        else:
            graph_params.append('rankdir=LR')

        block = '\n'.join(self.lines).replace('\n', '\n  ')
        text = TMPL.DOT.format(
            title=title,
            block=block,
            graph_params='\n  '.join(graph_params),
        ).replace('\n  \n', '\n\n')
        #print(text)
        return text


def generate_dot(gen: Generator, title: str, statements: model.Statements,
                 items_by_name: dict[str, model.Item]) -> str:
    """Iterate over statements and generate a dot source file"""
    def get_item(name: str) -> Optional[model.Item]:
        return None if name == '*' else items_by_name[name]

    for statement in statements:
        match statement:
            case model.Item() as item:
                gen.generate_item(item)

            case model.Connection() as connection:
                src_item = get_item(connection.src)
                dst_item = get_item(connection.dst)
                gen.generate_connection(connection, src_item, dst_item)

            case model.Style() as style:
                gen.generate_style(style)

            case model.Frame() as frame:
                gen.generate_frame(frame)

    return gen.generate_dot_text(title)


def filter_statements(statements: model.Statements) -> model.Statements:
    # collect used items
    connected_items = set()
    for statement in statements:
        match statement:
            case model.Connection() as conn: pass
            case _: continue
        for point in conn.src, conn.dst:
            connected_items.add(point)

    # filter out unconnected items
    new_statements = []
    for statement in statements:
        match statement:
            case model.Item() as item:
                if item.hidable and item.name not in connected_items:
                    continue
        new_statements.append(statement)

    return new_statements


def handle_options(statements: model.Statements) -> tuple[
        model.Statements, model.GraphOptions]:
    new_statements = []
    options = model.GraphOptions()
    for statement in statements:
        prefix = model.mk_err_prefix_from(statement.source)
        match statement:
            case model.Style() as style:
                match style.style:
                    case 'vertical':
                        options.is_vertical = True
                    case 'context':
                        options.is_context = True
                    case 'horizontal':
                        options.is_vertical = False
                    case 'item-text-width':
                        try:
                            options.item_text_width = int(style.value)
                        except ValueError as e:
                            raise model.DfdException(f'{prefix}{e}"')
                    case 'connection-text-width':
                        try:
                            options.connection_text_width = int(style.value)
                        except ValueError as e:
                            raise model.DfdException(f'{prefix}{e}"')

                    case _:
                        raise model.DfdException(f'{prefix}Unsupported style '
                                                 f'"{style.style}"')

                continue
        new_statements.append(statement)

    return new_statements, options



"""A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
"""

from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="data-flow-diagram",
    version="1.9.1",
    description="Commandline tool to generate data flow diagrams from text",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pbauermeister/dfd",
    author="Pascal Bauermeister",
    author_email="pascal.bauermeister@gmail.com",
    classifiers=[
        # https://pypi.org/classifiers/ :
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Information Technology",
        "Topic :: Software Development",
        "Topic :: Software Development :: Documentation",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
    ],
    keywords="diagram-generator, development, tool",
    license="GNU General Public License v3 (GPLv3)",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.10, <4",
    install_requires=[],
    extras_require={
        "dev": ["check-manifest"],
        "test": ["coverage"],
    },
    package_data={
#        "data_flow_diagram": ["tbdpackage__data.dat"],
    },
#    data_files=[('data_flow_diagram', ["VERSION"])],
    # The following would provide a command called `data-flow-diagram` which
    # executes the function `main` from this package when invoked:
    entry_points={
        "console_scripts": [
            "data-flow-diagram=data_flow_diagram:main",
        ],
    },
    project_urls={
        "Bug Reports": "https://github.com/pbauermeister/dfd/issues",
#        "Funding": "https://donate.pypi.org",
#        "Say Thanks!": "http://saythanks.io/to/example",
        "Source": "https://github.com/pbauermeister/dfd",
    },
)



"""Graphviz dot-related generation process"""

import subprocess
import sys

from . import model
from .error import print_error


def generate_image(graph_options: model.GraphOptions,
                   text: str, output_path: str, format: str) -> None:
    if graph_options.is_context:
        engine = 'neato'  # circo is not as good
    else:
        engine = 'dot'

    cmd = [engine, f'-T{format}',  f'-o{output_path}']
    try:
        subprocess.run(cmd, input=text, encoding='utf-8', check=True)
    except subprocess.CalledProcessError as e:
        for n, line in enumerate(text.splitlines()):
            print(f'{n+1:2}: {line}', file=sys.stderr)
        print_error(f'ERROR: {e}')
        sys.exit(1)
    #print('Generated:', output_path, file=sys.stderr)


def check_installed() -> None:
    cmd = ['dot', '-V']
    try:
        subprocess.run(cmd, stderr=subprocess.DEVNULL)
    except FileNotFoundError as e:
        print(f'ERROR: "Graphviz" seems not installed:', e, file=sys.stderr)
        sys.exit(2)
